import logging
import os
import random
import sys
from time import sleep

import peewee
import redis
from playhouse.db_url import connect

from telstar.com import Message
from telstar.consumer import MultiConsumer

link = redis.from_url(os.environ["REDIS"])
db = connect(os.environ["DATABASE"])
db.connect()

logger = logging.getLogger('telstar')
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.DEBUG)


class Test(peewee.Model):
    number = peewee.IntegerField()
    group_name = peewee.CharField()
    topic = peewee.CharField()

    class Meta:
        indexes = (
            (('number', 'group_name', 'topic'), True),
        )
        database = db


if __name__ == "__main__":

    if len(sys.argv) > 1 and sys.argv[1] == "setup":
        print("Recreating table in order to start from scratch")
        db.drop_tables([Test])
        db.create_tables([Test])

    def simple(consumer, record: Message, done):
        with db.atomic():
            Test.create(number=int(record.data["value"]), group_name=consumer.group_name, topic=record.stream)
            sleep(random.randrange(int(os.environ.get("SLEEPINESS"))) / 100)
            done()

    MultiConsumer(link=link,
                  group_name=os.environ.get("GROUP_NAME"),
                  consumer_name=os.environ.get("CONSUMER_NAME"),
                  config={
                      os.environ["STREAM_NAME"]: simple,
                      os.environ["STREAM_NAME_TWO"]: simple
                  }).run()



import json
import uuid
from datetime import datetime
from typing import Dict, Union

import peewee
from peewee import ModelSelect


class MessageError(Exception):
    pass


class TelstarEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, datetime):
            return o.isoformat()

        if isinstance(o, uuid.UUID):
            return str(o)

        return json.JSONEncoder.default(self, o)


class JSONField(peewee.TextField):
    def db_value(self, value: Dict[str, Union[int, str]]) -> str:
        return json.dumps(value, cls=TelstarEncoder)

    def python_value(self, value: str) -> Dict[str, Union[int, str]]:
        if value is not None:
            return json.loads(value)


class Message(object):
    IDFieldName = b"message_id"
    DataFieldName = b"data"

    def __init__(self, stream: str, msg_uuid: uuid.UUID, data: dict) -> None:
        if not isinstance(msg_uuid, uuid.UUID):
            raise TypeError(f"msg_uuid needs to be uuid.UUID not {type(msg_uuid)}")
        if isinstance(stream, bytes):
            stream = stream.decode("ascii")
        self.stream = stream.replace("telstar:stream:", "")
        self.msg_uuid = msg_uuid
        self.data = data

    def __repr__(self):
        return f"<Message self.stream:{self.stream} msd_id:{self.msg_uuid} data:{self.data}>"


class StagedMessage(peewee.Model):
    msg_uid = peewee.UUIDField(default=uuid.uuid4, index=True)
    topic = peewee.CharField(index=True)
    data = JSONField()

    sent = peewee.BooleanField(default=False, index=True)
    created_at = peewee.TimestampField(resolution=10**3)

    @classmethod
    def unsent(cls) -> ModelSelect:
        return cls.select().where(cls.sent == False)  # noqa

    def to_msg(self) -> Message:
        return Message(self.topic, self.msg_uid, self.data)


def increment_msg_id(id) -> bytes:
    # IDs are of the form "1509473251518-0" and comprise a millisecond
    # timestamp plus a sequence number to differentiate within the timestamp.
    time, sequence = id.decode("ascii").split("-")
    if not sequence:
        raise Exception("Argument error, {id} has wrong format not #-#")
    next_sequence = int(sequence) + 1

    return bytes(f"{time}-{next_sequence}", "ascii")


def decrement_msg_id(id: bytes) -> bytes:
    time, sequence = id.decode("ascii").split("-")
    if not sequence:
        raise Exception("Argument error, {id} has wrong format not #-#")
    sequence = int(sequence)
    time = int(time)
    if sequence == 0:
        time = time - 1
    else:
        sequence = int(sequence) - 1

    return bytes(f"{time}-{sequence}", "ascii")



import json
import logging
import os
import random
import sys
import uuid
from time import sleep
from typing import Callable, List, Tuple

import peewee
import redis
from playhouse.db_url import connect

from telstar.com import Message
from telstar.producer import Producer

link = redis.from_url(os.environ["REDIS"])
db = connect(os.environ["DATABASE"])
db.connect()

logger = logging.getLogger('telstar')
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.DEBUG)


class JSONField(peewee.TextField):
    def db_value(self, value):
        return json.dumps(value)

    def python_value(self, value):
        if value is not None:
            return json.loads(value)


class Events(peewee.Model):
    msg_uid = peewee.UUIDField(default=uuid.uuid4)
    topic = peewee.CharField()
    data = JSONField()

    class Meta:
        database = db


if __name__ == "__main__":

    if len(sys.argv) > 1:
        if sys.argv[1] == "setup":
            print("Recreating table in order to start from scratch")
            db.drop_tables([Events])
            db.create_tables([Events])

        if "create" in sys.argv[1:]:
            for i in range(int(os.environ["RANGE_FROM"]), int(os.environ["RANGE_TO"])):
                Events.create(topic=os.environ["STREAM_NAME"], data=dict(value=i))

    def puller() -> Tuple[List[Message], Callable[[], None]]:
        qs = Events.select().order_by(Events.id).limit(5)
        msgs = [Message(e.topic, e.msg_uid, e.data) for e in qs]

        def done():
            if not os.environ.get("KEEPEVENTS"):
                for r in qs:
                    r.delete_instance()
            sleep(random.randrange(int(os.environ.get("SLEEPINESS"))) / 10)
        return msgs, done
    print("starting")
    Producer(link, puller, context_callable=db.atomic).run()



import json
import uuid
from typing import Dict, List, Optional, Tuple, Union

import redis

from .com import Message, decrement_msg_id


class admin:
    def __init__(self, link: redis.Redis) -> None:
        self.link: redis.Redis = link

    def get_streams(self, match: None = None) -> List["Stream"]:
        match = match or ""
        # We wanted to use `scan_iter` but for an yet unknown reason
        # `SCAN` take hundreds of iterations to find all streams.
        streams = self.link.keys(f"telstar:stream:{match}*")
        return [Stream(self, s) for s in streams]

    def get_consumers(self) -> List["Consumer"]:
        return sum([g.get_consumers() for s in self.get_streams() for g in s.get_groups()], [])


class Stream:
    def __init__(self, admin: admin, stream_name: str) -> None:
        self.name = stream_name
        self.admin = admin
        self.link = admin.link

    @property
    def display_name(self) -> bytes:
        return self.name.replace(b"telstar:stream:", b"")

    def get_groups(self) -> List["Group"]:
        return [Group(self, name=info["name"], **self.link.xpending(self.name, info["name"]))
                for info in self.link.xinfo_groups(self.name)]

    def get_pending_messages(self) -> List["AdminMessage"]:
        return sum([g.get_pending_messages() for g in self.get_groups()], [])

    def get_length(self) -> int:
        return self.link.xlen(self.name)


class Group:
    def __init__(self, stream: Stream, name: str, pending: int, min: Optional[bytes], max: Optional[bytes], consumers: List[Dict[str, Union[bytes, int]]]) -> None:
        self.stream = stream
        self.link = stream.link
        self.name = name
        self.pending, self.min, self.max, self.consumers = pending, min, max, consumers

    def get_pending_messages(self) -> List["AdminMessage"]:
        if self.pending == 0:
            return []
        return [AdminMessage(self, **info)
                for info in self.link.xpending_range(self.stream.name, self.name, self.min, self.max, self.pending)]

    def get_consumers(self) -> List["Consumer"]:
        return [Consumer(self, **info) for info in self.link.xinfo_consumers(self.stream.name, self.name)]

    def get_seen_messages(self) -> int:
        stream_name = self.stream.name.replace(b"telstar:stream:", b"").decode("ascii")
        name = self.name.decode("ascii")
        return len(self.link.keys(f"telstar:seen:{stream_name}:{name}*"))

    def delete(self) -> bool:
        return self.link.xgroup_destroy(self.stream.name, self.name)


class Consumer:
    def __init__(self, group: Group, name: bytes, pending: int, idle: int) -> None:
        self.group = group
        self.name = name
        self.pending_messages = pending
        self.idle_time = idle

    def delete(self) -> int:
        return self.group.stream.admin.link.xgroup_delconsumer(self.group.stream.name, self.group.name, self.name)


class AdminMessage:
    def __init__(self, group: Group, message_id: bytes, consumer: str, time_since_delivered: int, times_delivered: int) -> None:
        self.group = group
        self.message_id = message_id
        self.consumer = consumer
        self.time_since_delivered = time_since_delivered
        self.times_delivered = times_delivered

    def read_raw(self) -> List[List[Union[bytes, List[Tuple[bytes, Dict[bytes, bytes]]]]]]:
        return self.group.stream.admin.link.xread({
            self.group.stream.name: decrement_msg_id(self.message_id)
        }, count=1)

    def read(self) -> Message:
        for stream_name, records in self.read_raw():
            for record in records:
                stream_msg_id, record = record
                return Message(stream_name,
                               uuid.UUID(record[Message.IDFieldName].decode("ascii")),
                               json.loads(record[Message.DataFieldName]))



import json
import logging
import threading
import time
import uuid
from functools import partial
from typing import Callable, Dict

import redis

from .com import Message, decrement_msg_id, increment_msg_id, MessageError

# An important concept to understand here is the consumer group which give us the following consumer properties:
# msg   -> consumer
#
# msg:1 -> cg-userSignUp.1
# msg:2 -> cg-userSignUp.2
# msg:3 -> cg-userSignUp.3
# msg:4 -> cg-userSignUp.1
# msg:5 -> cg-userSignUp.2
# msg:6 -> cg-userSignUp.3
# msg:7 -> cg-userSignUp.1
#
# Which basically means that inside a group a single consumer will only get message the others have not yet seen.
# For a deep dive read to this -> https://redis.io/topics/streams-intro
# This allows us to create N consumers w/o needing to figure out which message already has been processed.

log = logging.getLogger(__name__)


class MultiConsumer(object):

    def __init__(self, link: redis.Redis, group_name: str, consumer_name: str, config: dict, block: int = 2000, claim_the_dead_after: int = 20 * 1000, error_handlers=None) -> None:
        self.link = link
        self.block = block
        self.claim_the_dead_after = claim_the_dead_after
        self.consumer_name = consumer_name
        self.group_name = group_name
        self.error_handlers = error_handlers or {}

        self.processors = {f"telstar:stream:{stream_name}": fn
                           for stream_name, fn in config.items()}

        self.streams = self.processors.keys()
        for stream_name in self.streams:
            self.create_consumer_group(stream_name)

    def get_consumer_name(self, stream: str) -> str:
        return f"cg:{self.group_name}:{self.consumer_name}"

    def _seen_key(self, msg: Message) -> str:
        return f"telstar:seen:{msg.stream}:{self.group_name}:{msg.msg_uuid}"

    def _checkpoint_key(self, stream: str) -> str:
        return f"telstar:checkpoint:{stream}:{self.get_consumer_name(stream)}"

    # A new consumer group for the given stream, if the stream does not exist yet
    # create one (`mkstream`) - if it does we want all messages present `id=0`
    def create_consumer_group(self, stream_name: str) -> None:
        try:
            self.link.xgroup_create(stream_name, self.group_name, mkstream=True, id="0")
        except redis.exceptions.ResponseError:
            log.debug(f"Group: {self.group_name} for Stream: {stream_name} already exists")

    # In consumer groups, consumers can disappear, when they do they can leave non ack'ed message
    # which we want to claim and be delivered to a new consumer
    def claim_message_from_the_dead(self, stream_name: str) -> None:
        # Get information about all consumers in the group and how many messages are pending
        pending_info = self.link.xpending(stream_name, self.group_name)
        # {'pending': 10,
        #  'min': b'1560032216285-0',
        #  'max': b'1560032942270-0',
        #  'consumers': [{'name': b'cg-userSignUp.1', 'pending': 10}]}
        # Nothing to do
        if pending_info["pending"] == 0:
            log.debug(f"Stream: {stream_name} in Group: {self.group_name} has no pending messages")
            return
        # Get all messages ids within that range and select the ones we want to claim and claim them
        # But only if they are pending for more than 20secs.
        pending_messages = self.link.xpending_range(stream_name, self.group_name,
                                                    pending_info["min"], pending_info["max"], pending_info["pending"])
        # [
        #   {'message_id': b'1560194528886-0',
        #    'consumer': b'cg-userSignUp.1',
        #    'time_since_delivered': 22020,
        #    'times_delivered': 1}
        #  ...]
        messages_to_claim = [p["message_id"] for p in pending_messages]

        if not messages_to_claim:
            # The pending messages are all our own no need to claim anything
            # This can happen when we simply restart a consumer with the same name
            return

        # It might be cheaper to claim *and* receive the message so we can work on them directly
        # w/o catching up through the history with the potential of a lot of already seen keys.
        log.debug(f"Stream: {stream_name} in Group: {self.group_name} claiming: {len(messages_to_claim)} message(s)")
        claimed_messages = self.link.xclaim(stream_name, self.group_name, self.consumer_name, self.claim_the_dead_after, messages_to_claim, justid=True)
        log.debug(f"Stream: {stream_name} in Group: {self.group_name} claimed: {len(messages_to_claim)} message(s)")
        return claimed_messages

    # We claim the message from other dead/non-responsive consumers.
    # When new message have been claimed they are usually from the past
    # which means in order to process them we need to start processing our history.
    def transfer_and_process_stream_history(self, streams: list):
        last_seen = dict()
        for stream_name in streams:
            last_seen[stream_name] = self.get_last_seen_id(stream_name)
            stream_msg_ids = self.claim_message_from_the_dead(stream_name)
            if stream_msg_ids:
                # if there are message that we have claimed we need to determine where to start processing
                # because we can't just wait for new message to arrive.
                before_earliest = decrement_msg_id(min(stream_msg_ids))
                next_after_seen = increment_msg_id(last_seen[stream_name])
                last_seen[stream_name] = min([before_earliest, next_after_seen])
        # Read all message for the past up until now.
        log.info(f"Stream: {', '.join(last_seen)} in Group: {self.group_name} as Consumer: {self.consumer_name} reading past messages")
        self.catchup(last_seen)

    # This is the main loop where we start from the history
    # and claim message and reprocess our history.
    # We also loop the transfer_and_process_history as other consumers might have died while we waited
    def run(self):
        log.info(f"Starting consumer loop for Group {self.group_name}")
        while True:
            self.run_once()

    def run_once(self) -> None:
        self.transfer_and_process_stream_history(self.streams)
        # With our history processes we can now start waiting for new message to arrive `>`
        config = {k: ">" for k in self.streams}
        log.info(f"Stream: {', '.join(self.streams)} in Group: {self.group_name} as Consumer: {self.consumer_name} reading pending message or waiting for new")
        self.read(config, block=self.block)

    def get_last_seen_id(self, stream_name: str) -> bytes:
        check_point_key = self._checkpoint_key(stream_name)
        return self.link.get(check_point_key) or b"0-0"

    # Multiple things are happening here.
    # 1. Save the stream_msg_id as checkpoint, which means
    #    that we know where to start should the consumer be restarted
    # 2. Each message has a UUID and in order to process each meassage only once we remember
    #    the UUID for 14 days
    # 3. Acknowledge the message to meaning that we have processed it
    def acknowledge(self, msg: Message, stream_msg_id: bytes) -> None:
        log.debug(f"Stream: telstar:stream:{msg.stream} in Group: {self.group_name} acknowledging Message: {msg.msg_uuid} - {stream_msg_id}")
        check_point_key = self._checkpoint_key(f"telstar:stream:{msg.stream}")
        seen_key = self._seen_key(msg)
        # Execute the following statments in a transaction e.g. redis speak `pipeline`
        pipe = self.link.pipeline()

        # If this key changes before we execute the pipeline than the ack fails and this the processor reverts all the work.
        # Which is exactly what we want in this case as the work has already been completed by another consumer.
        pipe.watch(seen_key)

        # Mark this as a seen key for 14 Days meaning if the message reappears after 14 days we reprocess it
        pipe.set(seen_key, 1, ex=14 * 24 * 60 * 60)  # 14 days

        # Set the checkpoint for this consumer so that it knows where to start agains once it restarts.
        pipe.set(check_point_key, stream_msg_id)

        # Acknowledge the actual message
        pipe.xack(f"telstar:stream:{msg.stream}", self.group_name, stream_msg_id)
        pipe.execute()

    def work(self, stream_name: bytes, stream_msg_id: bytes, record: Dict[bytes, bytes]) -> None:
        try:
            msg = Message(stream_name,
                          uuid.UUID(record[Message.IDFieldName].decode("ascii")),
                          json.loads(record[Message.DataFieldName]))
        except KeyError as exc:
            msg = f"Malformed message, record: {record} does not have fields {Message.IDFieldName} and {Message.DataFieldName} "
            log.exception(msg)
            raise MessageError(msg) from exc

        done = partial(self.acknowledge, msg, stream_msg_id)
        key = self._seen_key(msg)
        if self.link.get(key):
            # This is a double send
            log.debug(f"Stream: telstar:stream:{msg.stream} in Group: {self.group_name} skipping already processed Message: {msg.msg_uuid} - {stream_msg_id} ")
            return done()

        log.info(f"Stream: telstar:stream:{msg.stream} in Group: {self.group_name} processing Message: {msg.msg_uuid} - {stream_msg_id}")
        self.processors[stream_name.decode("ascii")](self, msg, done)

    # Process all message from `start`
    def catchup(self, streams: Dict[str, bytes]) -> int:
        return self._xreadgroup(streams)

    # Process wait for new messages
    def read(self, streams: Dict[str, str], block: int) -> int:
        return self._xreadgroup(streams, block=block)

    def _xreadgroup(self, streams: Dict[str, str], block: int = 0) -> int:
        result = list()
        for stream_name, records in self.link.xreadgroup(self.group_name, self.consumer_name, streams, block=block):
            for record in records:
                stream_msg_id, record = record
                result.append((stream_name, stream_msg_id, record))

        if not result:
            return 0
        # Sort the message afterwards in order to restore the order they where sent in, this can only be a best effort
        # approach and does not guarantee the correct order when using `xreadgroup` with multiple streams.
        for processed, t in enumerate(sorted(result, key=lambda t: t[1]), start=1):
            stream_name, stream_msg_id, record = t
            try:
                self.work(stream_name, stream_msg_id, record)
            except Exception as exc:
                self._handle_exception(exc, stream_name, stream_msg_id, record)
        return processed

    def _find_error_handler(self, exc):
        for cls in type(exc).__mro__:
            handler = self.error_handlers.get(cls)
            if handler is not None:
                return handler

    def _handle_exception(self, exc, stream_name, stream_msg_id, record):
        handler = self._find_error_handler(exc)

        if handler is None:
            raise exc

        bare_ack = partial(self._bare_ack, stream_name, stream_msg_id)
        return handler(exc, bare_ack)

    def _bare_ack(self, stream_name, stream_msg_id):
        check_point_key = self._checkpoint_key(stream_name)
        pipe = self.link.pipeline()

        pipe.set(check_point_key, stream_msg_id)
        pipe.xack(stream_name, self.group_name, stream_msg_id)
        pipe.execute()


class Consumer(MultiConsumer):
    def __init__(self, link: redis.Redis, group_name: str, consumer_name: str, stream_name: str, processor_fn: Callable) -> None:
        super().__init__(link, group_name, consumer_name, {stream_name: processor_fn})


class MultiConsumeOnce(MultiConsumer):
    # This lets you write code that is executed once against an entire stream
    # TODO: What would be really cool is to have this sort of like migrations
    #       where `telstar` itself has cli commands to add and run files containing
    #       `MultiConsumeOnce` code.
    def __init__(self, link: redis.Redis, group_name: str, config: dict) -> None:
        super().__init__(link, group_name, "once-consumer", config, 2000, 20000)

    def _applied_key(self) -> str:
        return f"telstar:once:{self.group_name}"

    def is_applied(self) -> bool:
        key = self._applied_key()
        return bool(self.link.get(key))

    def mark_as_applied(self) -> bool:
        key = self._applied_key()
        return self.link.set(key, int(time.time()))

    def has_pending_message(self) -> bool:
        for stream_name in self.streams:
            if self.link.xpending(stream_name, self.group_name)["pending"] != 0:
                return True
        return False

    def run(self) -> int:
        num_processed = 0
        if self.is_applied():
            log.info(f"Group: {self.group_name} for Streams: {self.streams} will not run as it already ran")
            return num_processed

        # This is the first time we try to apply this.
        if not self.has_pending_message():
            # Reading a stream from ">" has a special meaning, it instructs redis to send all messages to the group
            # Which does two things first it puts them all into the pending list of that consumer inside the group
            # and also delivers them to the client.
            streams = {s: ">" for s in self.streams}
            num_processed = self.read(streams, 0)
        else:
            # Now the data as already been delivered to the group we can now start reading from the beginning
            streams = {s: "0" for s in self.streams}
            num_processed = self.read(streams, 0)

        # everything has been seen and processed where can mark this as applied
        if not self.has_pending_message():
            self.mark_as_applied()
        return num_processed


class PropagatingThread(threading.Thread):
    def run(self):
        self.exc = None
        try:
            self.ret = self._target(*self._args, **self._kwargs)
        except BaseException as e:
            self.exc = e

    def join(self) -> None:
        super(PropagatingThread, self).join()
        if self.exc:
            raise self.exc
        return self.ret


class ThreadedMultiConsumer:
    def __init__(self, link: redis.Redis, consumer_name: str, group_configs: dict, **kw) -> None:
        self.consumers = list()
        for group_name, config in group_configs.items():
            self.consumers.append(MultiConsumer(link, group_name, consumer_name, config, **kw))

    def run(self):
        self._run_threaded("run")

    def run_once(self) -> None:
        self._run_threaded("run_once")

    def _run_threaded(self, target: str) -> None:
        threads = list()
        for c in self.consumers:
            t = PropagatingThread(target=getattr(c, target), daemon=True)
            t.start()
            threads.append(t)

        for t in threads:
            t.join()



import os
import uuid
from datetime import datetime
from unittest import mock

import pytest
import redis
import peewee
from playhouse.db_url import connect

import telstar
from telstar.com import Message, StagedMessage, MessageError
from telstar.consumer import Consumer, MultiConsumer, MultiConsumeOnce
from telstar.producer import StagedProducer
from marshmallow import fields, Schema, ValidationError


@pytest.fixture
def msg_schema() -> Schema:
    class MyObjSchema(Schema):
        name = fields.Str()
        email = fields.Email()
    return MyObjSchema


@pytest.fixture
def link() -> redis.Redis:
    return mock.MagicMock(spec=redis.Redis)


@pytest.fixture
def reallink() -> redis.Redis:
    client = redis.from_url(os.environ.get("REDIS", "redis://localhost:6379/10"))
    client.flushdb()
    return client


@pytest.fixture
def realdb() -> peewee.Database:
    db = connect(os.environ.get("DATABASE", "postgres://127.0.0.1:5432/telstar-integration-test"))
    tables = [StagedMessage]
    db.bind(tables)
    db.drop_tables(tables)
    db.create_tables(tables)
    return db


@pytest.fixture
def db() -> peewee.Database:
    db = connect("sqlite:///:memory:")
    db.bind([StagedMessage])
    db.create_tables([StagedMessage])
    return db


@pytest.fixture
def consumer(link) -> Consumer:
    return Consumer(link, "mygroup", "myname", "mytopic", lambda msg, done: done())


def test_message():
    uid = uuid.uuid4()
    m = Message("topic", uid, dict())
    assert m.msg_uuid == uid


def test_consumer_create_group(link):
    Consumer(link, "mygroup", "myname", "mytopic", lambda msg, done: done())
    link.xgroup_create.assert_called_once_with('telstar:stream:mytopic', 'mygroup', id='0', mkstream=True)


def test_consumer_run(link: redis.Redis):
    callback = mock.Mock()
    msg_id = str(uuid.uuid4()).encode("ascii")
    link.get.return_value = None
    link.xreadgroup.return_value = [[
        b"telstar:stream:mytopic", [["stream_msg_id", {b'message_id': msg_id, b"data": "{}"}]]
    ]]
    c = Consumer(link, "mygroup", "myname", "mytopic", callback)
    c.transfer_and_process_stream_history = lambda *a, **kw: None
    c.run_once()
    callback.assert_called()


def test_consumer_run_callback(link: redis.Redis):
    called = False
    msg_id: bytes = str(uuid.uuid4()).encode("ascii")

    def callback(c, msg: Message, done):
        nonlocal called, msg_id
        called = True
        assert msg.msg_uuid == uuid.UUID(msg_id.decode("ascii"))
        assert msg.data == {}
        assert msg.stream == "mytopic"
        done()

    link.get.return_value = None
    link.xreadgroup.return_value = [[
        b"telstar:stream:mytopic", [["stream_msg_id", {b'message_id': msg_id, b"data": "{}"}]]
    ]]
    c = Consumer(link, "mygroup", "myname", "mytopic", callback)
    c.transfer_and_process_stream_history = lambda *a, **kw: None
    c.run_once()

    assert called is True


def test_consumer_checkpoint(link: redis.Redis):
    msg_id: bytes = b"49ccc393-1594-4d06-b357-0e1f322300b2"

    def callback(c, msg: Message, done):
        done()

    link.get.return_value = None
    pipeline = mock.MagicMock(spec=redis.client.Pipeline)()
    link.pipeline.return_value = pipeline
    link.xreadgroup.return_value = [[
        b"telstar:stream:mytopic", [["stream_msg_id", {b'message_id': msg_id, b"data": "{}"}]]
    ]]
    c = Consumer(link, "mygroup", "myname", "mytopic", callback)
    c.run_once()
    link.get.assert_any_call("telstar:checkpoint:telstar:stream:mytopic:cg:mygroup:myname")
    pipeline.set.assert_called_with("telstar:checkpoint:telstar:stream:mytopic:cg:mygroup:myname", "stream_msg_id")


def test_consumer_with_multiple_stearms(link):
    callback1 = mock.Mock()
    callback2 = mock.Mock()

    config = {
        "mytopic1": callback1,
        "mytopic2": callback2
    }
    msg_id = str(uuid.uuid4()).encode("ascii")
    link.get.return_value = None
    link.xreadgroup.return_value = [
        [b"telstar:stream:mytopic1", [
            ["stream_msg_id1", {b'message_id': msg_id, b"data": "{}"}],
            ["stream_msg_id2", {b'message_id': msg_id, b"data": "{}"}]
        ]],

        [b"telstar:stream:mytopic2", [["stream_msg_id", {b'message_id': msg_id, b"data": "{}"}]]]
    ]

    mc = MultiConsumer(link, "group", "name", config)
    mc.transfer_and_process_stream_history = lambda *a, **kw: None
    mc.run_once()

    assert callback1.call_count == 2
    assert callback2.call_count == 1


def test_seen_key(consumer: Consumer):
    uid_hex = "752884c3f7284cf19d3b9940373685f4"
    uid = uuid.UUID(uid_hex)
    m = Message("mytopic", uid, dict())
    assert consumer._seen_key(m) == "telstar:seen:mytopic:mygroup:752884c3-f728-4cf1-9d3b-9940373685f4"


def test_seen_key_bytes(consumer: Consumer):
    uid_hex = "752884c3f7284cf19d3b9940373685f4"
    uid = uuid.UUID(uid_hex)
    m = Message(b"mytopic", uid, dict())
    assert consumer._seen_key(m) == "telstar:seen:mytopic:mygroup:752884c3-f728-4cf1-9d3b-9940373685f4"


def test_message_strip_telstar_prefix(consumer: Consumer):
    uid_hex = "752884c3f7284cf19d3b9940373685f4"
    uid = uuid.UUID(uid_hex)
    m = Message(b"telstar:stream:mytopic", uid, dict())
    assert consumer._seen_key(m) == "telstar:seen:mytopic:mygroup:752884c3-f728-4cf1-9d3b-9940373685f4"


def test_message_with_non_uuid():
    uid = "something random"
    with pytest.raises(TypeError):
        Message("mytopic", uid, dict())


def test_checkpoint_key(consumer: Consumer):
    assert consumer._checkpoint_key("mytopic") == "telstar:checkpoint:mytopic:cg:mygroup:myname"


def test_staged_event(db):
    telstar.stage("mytopic", dict(a=1))
    assert len(StagedMessage.select().where(StagedMessage.topic == "mytopic")) == 1


def test_staged_producer(db, link):
    telstar.stage("mytopic", dict(a=1))
    [msgs], _ = StagedProducer(link, db).get_records()
    assert msgs.stream == "mytopic"
    assert msgs.data == dict(a=1)


def test_encoding_raises_correct_type_error(db, link):
    now = datetime.now()
    with pytest.raises(TypeError):
        telstar.stage("mytopic", dict(dt=now, mock=mock.MagicMock()))


def test_stage_can_encode_types(db, link):
    now = datetime.now()
    uid = uuid.uuid4()
    telstar.stage("mytopic", dict(dt=now, uuid=uid))
    [msg], cb = StagedProducer(link, db).get_records()
    assert msg.data == {"dt": now.isoformat(), "uuid": str(uid)}


def test_staged_producer_done_callback_removes_staged_events(db, link):
    telstar.stage("mytopic", dict(a=1))
    telstar.stage("mytopic", dict(a=1))
    msgs, cb = StagedProducer(link, db, batch_size=1).get_records()
    assert len(msgs) == 1
    assert len(telstar.staged()) == 2
    cb()
    msgs, _ = StagedProducer(link, db).get_records()
    assert len(msgs) == 1
    assert len(telstar.staged()) == 1


def test_consumer_once_keys(link):
    callback = mock.Mock()
    m = MultiConsumeOnce(link, "testgroup", {"mystream": callback})
    assert m._applied_key() == "telstar:once:testgroup"


@pytest.mark.integration
def test_app_pattern(realdb, reallink, msg_schema):
    app = telstar.app(reallink, consumer_name="c1")
    m = mock.Mock()

    @app.consumer("group", ["mytopic", "mytopic2"], schema=msg_schema)
    def callback(data: dict):
        m()

    telstar.stage("mytopic", dict(name="1", email="a@b.com"))
    telstar.stage("mytopic2", dict(name="1", email="a@b.com"))
    StagedProducer(reallink, realdb).run_once()

    app.run_once()
    assert m.call_count == 2


@pytest.mark.integration
def test_app_consumer_strictness(realdb, reallink, msg_schema):
    app = telstar.app(reallink, consumer_name="c1")

    @app.consumer("group", "mytopic", schema=msg_schema, strict=True)
    def callback(data: dict):
        print(data)

    telstar.stage("mytopic", dict(name="1", email="invalid"))
    StagedProducer(reallink, realdb).run_once()

    with pytest.raises(ValidationError):
        app.run_once()


@pytest.mark.integration
def test_app_consumer_errorhandler(realdb, reallink, msg_schema):
    app = telstar.app(reallink, consumer_name="c1")
    m = mock.Mock()

    @app.consumer("group", "mytopic", schema=msg_schema, strict=True)
    def callback(data: dict):
        print(data)

    @app.errorhandler(ValidationError)
    def handler(exc, ack):
        m()

    telstar.stage("mytopic", dict(name="1", email="invalid"))
    StagedProducer(reallink, realdb).run_once()

    app.run_once()
    assert m.call_count == 1


@pytest.mark.integration
def test_app_consumer_errorhandler_can_acknowledge(realdb, reallink, msg_schema):
    app = telstar.app(reallink, consumer_name="c1")
    m = mock.Mock()

    @app.consumer("group", "mytopic", schema=msg_schema, strict=True)
    def callback(data: dict):
        print(data)

    @app.errorhandler(ValidationError)
    def handler(exc, ack):
        m()
        ack()

    telstar.stage("mytopic", dict(name="1", email="invalid"))
    StagedProducer(reallink, realdb).run_once()

    app.run_once()
    app.run_once()
    assert m.call_count == 1  # The message is ack'ed in the error handler


@pytest.mark.integration
def test_app_consumer_invalid_message(realdb, reallink: redis.Redis, msg_schema):
    app = telstar.app(reallink, consumer_name="c1")
    m = mock.Mock()

    @app.consumer("group", "mytopic", schema=msg_schema, strict=True)
    def callback(data: dict):
        print(data)

    @app.errorhandler(MessageError)
    def handler(exc, ack):
        m()
        ack()

    reallink.xadd("telstar:stream:mytopic", {"not_uuid": "asd", "not_data": "asd"})
    app.run_once()
    app.run_once()
    assert m.call_count == 1  # The message is ack'ed in the error handler


@pytest.mark.integration
def test_app_consumer_ack_invalid(realdb, reallink, mocker, msg_schema):
    app = telstar.app(reallink, consumer_name="c1")

    @app.consumer("group", "mytopic", schema=msg_schema, acknowledge_invalid=True, strict=False)
    def callback(data: dict):
        pass

    telstar.stage("mytopic", dict(name="1", email="a@b.com"))
    StagedProducer(reallink, realdb).run_once()

    ack = mocker.spy(MultiConsumer, "acknowledge")
    app.run_once()
    ack.assert_called_once()


@pytest.mark.integration
def test_app_consumer_do_not_ack_invalid(realdb, reallink, mocker, msg_schema):
    app = telstar.app(reallink, consumer_name="c1")
    m = mock.Mock()

    @app.consumer("group", "mytopic", schema=msg_schema, acknowledge_invalid=False, strict=False)
    def callback(data: dict):
        m()

    telstar.stage("mytopic", dict(name="1", email="invalid"))
    StagedProducer(reallink, realdb).run_once()

    ack = mocker.spy(MultiConsumer, "acknowledge")
    app.run_once()
    app.run_once()
    ack.assert_not_called()


@pytest.mark.integration
def test_app_consumer_full_message(realdb, reallink, mocker, msg_schema):
    app = telstar.app(reallink, consumer_name="c1")
    m = mock.Mock()

    @app.consumer("group", "mytopic", schema=msg_schema, acknowledge_invalid=False, strict=False)
    def callback(data: Message):
        m(data)

    telstar.stage("mytopic", dict(name="1", email="a@b.com"))
    StagedProducer(reallink, realdb).run_once()

    app.run_once()
    [msg, ] = m.call_args[0]
    assert msg.data == dict(name="1", email="a@b.com")
    assert isinstance(msg, Message)


@pytest.mark.integration
def test_consumer_once(realdb, reallink):
    result = list()
    for i in range(10):
        telstar.stage("mytopic", dict(i=i))

    def callback(c, msg: Message, done):
        data = int(msg.data["i"])
        if data < c.stop:
            result.append(data)
            done()

    sp = StagedProducer(reallink, realdb, batch_size=100)
    sp.run_once()

    m = MultiConsumeOnce(reallink, "mytest", {"mytopic": callback})
    m.stop = 5
    assert m.run() == 10  # Successfully proccessed 10 message but only five got ack'ed

    m.stop = 10
    assert m.run() == 5  # Processes the remaining five
    assert m.run() == 0  # Nothing to process anylonger

    assert result == list(range(10))


@pytest.mark.integration
def test_admin_basics(reallink, realdb, msg_schema):
    app = telstar.app(reallink, consumer_name="c1")
    admin = telstar.admin(reallink)
    telstar.stage("mytopic", dict(name="1", email="a@b.com"))

    sp = StagedProducer(reallink, realdb, batch_size=100)
    sp.run_once()

    [stream] = admin.get_streams()
    # We have not yet read something from the stream thus there is now group and no consumer
    assert stream.get_pending_messages() == []
    assert stream.get_groups() == []

    @app.consumer("group", "mytopic", schema=msg_schema, acknowledge_invalid=False, strict=False)
    def callback(data: Message):
        raise Exception("Wont't process")

    with pytest.raises(Exception):
        app.run_once()
    with pytest.raises(Exception):
        app.run_once()

    [msg] = stream.get_pending_messages()
    [grp] = stream.get_groups()
    [consumer] = grp.get_consumers()

    assert stream.get_length() == 1
    assert grp.get_seen_messages() == 0
    assert stream.display_name == b"mytopic"

    assert msg.times_delivered == 2
    assert consumer.idle_time < 1000


@pytest.mark.integration
def test_admin_group_deletion(reallink, realdb, msg_schema):
    app = telstar.app(reallink, consumer_name="c1")
    admin = telstar.admin(reallink)
    telstar.stage("mytopic", dict(name="1", email="a@b.com"))

    sp = StagedProducer(reallink, realdb, batch_size=100)
    sp.run_once()

    @app.consumer("group", "mytopic", schema=msg_schema)
    def callback(data: Message):
        pass

    app.run_once()

    [stream] = admin.get_streams()
    [grp] = stream.get_groups()

    grp.delete()

    assert stream.get_groups() == []


@pytest.mark.integration
def test_admin_read_pending_message(reallink, realdb, msg_schema):
    app = telstar.app(reallink, consumer_name="c1")
    admin = telstar.admin(reallink)
    data = dict(name="1", email="a@b.com")
    telstar.stage("mytopic", data)

    sp = StagedProducer(reallink, realdb, batch_size=100)
    sp.run_once()

    @app.consumer("group", "mytopic", schema=msg_schema, acknowledge_invalid=False, strict=False)
    def callback(data: Message):
        raise Exception("Wont't process")

    with pytest.raises(Exception):
        app.run_once()

    [stream] = admin.get_streams()
    [grp] = stream.get_groups()
    [msg] = grp.get_pending_messages()
    assert msg.read().data == data


@pytest.mark.integration
def test_admin_consumer_deletion(reallink, realdb, msg_schema):
    app = telstar.app(reallink, consumer_name="c1")
    admin = telstar.admin(reallink)
    telstar.stage("mytopic", dict(name="1", email="a@b.com"))

    sp = StagedProducer(reallink, realdb, batch_size=100)
    sp.run_once()

    @app.consumer("group", "mytopic", schema=msg_schema)
    def callback(data: Message):
        pass

    app.run_once()

    [stream] = admin.get_streams()
    [grp] = stream.get_groups()
    [consumer] = grp.get_consumers()

    consumer.delete()

    assert grp.get_consumers() == []


@pytest.mark.integration
def test_consume_order(realdb, reallink):
    result = list()
    telstar.stage("mytopic", dict(i=1))
    telstar.stage("mytopic2", dict(i=2))
    telstar.stage("mytopic2", dict(i=3))
    telstar.stage("mytopic", dict(i=4))
    telstar.stage("mytopic2", dict(i=5))
    telstar.stage("mytopic2", dict(i=6))
    telstar.stage("mytopic2", dict(i=7))
    telstar.stage("mytopic", dict(i=8))
    telstar.stage("mytopic", dict(i=9))
    telstar.stage("mytopic2", dict(i=10))
    telstar.stage("mytopic2", dict(i=11))
    telstar.stage("mytopic2", dict(i=12))

    def callback(c, msg: Message, done):
        data = int(msg.data["i"])
        result.append(data)
        done()

    sp = StagedProducer(reallink, realdb, batch_size=100)
    sp.run_once()
    m = MultiConsumeOnce(reallink, "mytest", {"mytopic": callback, "mytopic2": callback})
    m.run()

    def monotonicity(l):
        # Count the number of times the current element is one smaller than the next
        return sum([x + 1 == y for x, y in zip(l, l[1:])])

    # Maximum monotony
    assert monotonicity(result) >= 3



import json
import logging
from time import sleep
from typing import Callable, List, Optional, Tuple

from peewee import Database
from redis.client import Redis

from .com import Message, StagedMessage

log = logging.getLogger(__name__)


class Producer(object):
    def __init__(self, link: Redis, get_records: Callable[[], Tuple[List[Message], Callable[[], None]]], context_callable: Optional[Callable] = None) -> None:
        self.link = link
        self.get_records = get_records
        self.context_callable = context_callable

    def run_once(self) -> None:
        records, done = self.get_records()
        pipe = self.link.pipeline()
        for msg in records:
            # Why the sleep here? It helps with sorting the events on the receiving side.
            # But it also limits to amount of possible sends to under 1k messages per send.
            # Which for now seems acceptable.
            sleep(.001)
            pipe.xadd(f"telstar:stream:{msg.stream}", {
                      Message.IDFieldName: str(msg.msg_uuid),
                      Message.DataFieldName: json.dumps(msg.data)})
        pipe.execute()
        done()

    def run(self):
        log.info("Starting main producer loop")
        while True:
            if callable(self.context_callable):
                with self.context_callable():
                    self.run_once()
            else:
                self.run_once()


class StagedProducer(Producer):
    def __init__(self, link: Redis, database: Database, batch_size: int = 5, wait: float = 0.5) -> None:
        self.batch_size = batch_size
        self.wait = wait
        StagedMessage.bind(database)

        super().__init__(link, self.create_puller(), StagedMessage._meta.database.atomic)

    def create_puller(self) -> Callable:
        def puller() -> Tuple[List[Message], Callable[[], None]]:
            qs = StagedMessage.unsent().limit(self.batch_size).order_by(StagedMessage.id)
            msgs = [e.to_msg() for e in qs]
            log.debug(f"Found {len(msgs)} messages to be send")

            def done():
                ids = list(map(lambda l: l.id, qs))
                if ids:
                    log.debug(f"Attempting to mark {len(ids)} messages as being sent")
                    result = StagedMessage.update(sent=True).where(StagedMessage.id << ids).execute()
                    log.debug(f"Result was: {result}")
                sleep(self.wait)

            return msgs, done
        return puller



"""
Telstar is a package to write producer and consumers groups against redis streams.
"""
import inspect
import logging
from datetime import datetime
from functools import wraps
from typing import Callable, Dict, List, Union
from uuid import UUID

import redis
from marshmallow import Schema, ValidationError

from .admin import admin
from .com import Message, StagedMessage
from .consumer import MultiConsumer, ThreadedMultiConsumer

__version__ = "0.2.5"


logging.getLogger(__package__).addHandler(logging.NullHandler())
log = logging.getLogger(__package__)

admin = admin


def stage(topic: str, data: Dict[str, Union[int, str, datetime, UUID]]) -> UUID:
    e = StagedMessage.create(topic=topic, data=data)
    return e.msg_uid


def staged() -> List[Message]:
    return [e.to_msg() for e in StagedMessage.unsent()]


class app:
    def __init__(self, link: redis.Redis, consumer_name: str, consumer_cls: MultiConsumer = ThreadedMultiConsumer, **kwargs) -> None:
        self.link: redis = link
        self.config: dict = {}
        self.consumer_name: str = consumer_name
        self.consumer_cls: MultiConsumer = consumer_cls
        self.kwargs = kwargs
        self.error_handlers = {}

    def _register_error_handler(self, exc_class, fn):
        self.error_handlers[exc_class] = fn

    def errorhandler(self, exc_class):
        def decorator(fn):
            self._register_error_handler(exc_class, fn)
        return decorator

    def get_consumer(self) -> MultiConsumer:
        return self.consumer_cls(self.link, self.consumer_name, self.config, error_handlers=self.error_handlers, **self.kwargs)

    def start(self):
        self.get_consumer().run()

    def run_once(self) -> None:
        self.get_consumer().run_once()

    def requires_full_message(self, fn: Callable) -> bool:
        argsspec = inspect.getfullargspec(fn)
        arg = argsspec.args[0]
        return argsspec.annotations[arg] is Message

    def consumer(self, group: str, streams: list, schema: Schema, strict=True, acknowledge_invalid=False) -> Callable:
        def decorator(fn):
            fullmessage = self.requires_full_message(fn)
            nonlocal streams
            if not isinstance(streams, list):
                streams = [streams]
            for stream in streams:
                @wraps(fn)
                def actual_consumer(consumer: MultiConsumer, msg: Message, done: callable):
                    try:
                        msg.data = schema().load(msg.data)
                        fn(msg) if fullmessage else fn(msg.data)
                        done()
                    except ValidationError as err:
                        log.error(f"Unable to validate message: {msg}", exc_info=True)
                        if acknowledge_invalid:
                            done()
                        if strict:
                            raise err

                if group in self.config:
                    self.config[group][stream] = actual_consumer
                else:
                    self.config[group] = {stream: actual_consumer}

        return decorator
